import base64
from typing import Any

class HttpAuth:
    """HTTP authentication headers factory"""

    @staticmethod
    def basic(username: str, password: str) -> dict[str, str]:
        credentials = f"{username}:{password}"
        encoded_credentials = base64.b64encode(credentials.encode()).decode()
        return {"Authorization": f"Basic {encoded_credentials}"}

    @staticmethod
    def api_key(api_key: str) -> dict[str, str]:
        return {"API-Key": api_key}

    @staticmethod
    def bearer(token: str) -> dict[str, str]:
        return {"Authorization": f"Bearer {token}"}

    @staticmethod
    def user_id_and_token(user_id: str, access_token: str) -> dict[str, str]:
        return {"user_id": f"{user_id}", "access_token": f"{access_token}"}

    @staticmethod
    def client_id_and_token(client_id: str, access_token: str) -> dict[str, str]:
        return {"client_id": f"{client_id}", "access_token": f"{access_token}"}

    @staticmethod
    def custom(custom_auth_headers: dict[str, Any]) -> dict[str, str]:
        return custom_auth_headers

    @staticmethod
    def none() -> dict[str, str]:
        return {}


import pytest
import asyncio
from core_infinity_stones.core.utils import retry


def test_sync_retry_success_after_failures() -> None:
    attempts = {"count": 0}

    @retry(max_attempts=3, delay_in_seconds=0)
    def flaky() -> str:
        attempts["count"] += 1
        if attempts["count"] < 3:
            raise ValueError("Fail")
        return "Success"

    assert flaky() == "Success"
    assert attempts["count"] == 3


def test_sync_retry_raises_after_max_attempts() -> None:
    attempts = {"count": 0}

    @retry(max_attempts=2, delay_in_seconds=0)
    def always_fails() -> None:
        attempts["count"] += 1
        raise RuntimeError("Always fails")

    with pytest.raises(RuntimeError, match="Always fails"):
        always_fails()

    assert attempts["count"] == 2


async def test_async_retry_success_after_failures() -> None:
    attempts = {"count": 0}

    @retry(max_attempts=4, delay_in_seconds=0)
    async def async_flaky() -> str:
        attempts["count"] += 1
        if attempts["count"] < 4:
            raise Exception("Temporary failure")
        return "OK"

    result = await async_flaky()
    assert result == "OK"
    assert attempts["count"] == 4


test_sync_retry_raises_after_max_attempts()
test_sync_retry_success_after_failures()
asyncio.run(test_async_retry_success_after_failures())


from enum import StrEnum
from typing import Any, Optional
from pydantic import BaseModel
from typing_extensions import deprecated

class Severity(StrEnum):
    """Severity levels for errors and events."""

    CRITICAL = "CRITICAL"
    """Service-down, data loss, or security breaches. Immediate attention required"""

    HIGH = "HIGH"
    """Major functional issue that affects important workflows. Requires prompt resolution."""

    MEDIUM = "MEDIUM"
    """Degraded functionality or partial impact. Should be fixed soon."""

    LOW = "LOW"
    """Minor errors that don’t impact most users or core functionality."""

    INFO = "INFO"
    """Informational errors that don’t impact the app behavior (e.g. validation errors or user mistakes)."""


class LocalizedMessage(BaseModel):
    ar: str
    en: str


class HttpErrorPublicDetails(BaseModel):
    code: str
    status_code: int
    message: LocalizedMessage
    details: dict[Any, Any]


class HttpErrorDebugDetails(BaseModel):
    severity: Severity
    debug_code: str
    occurred_while: Optional[str]
    caused_by: Optional[str] = None
    debug_message: str
    debug_details: dict[Any, Any] = {}

    class Config:
        arbitrary_types_allowed = True
        json_encoders = {
            Exception: lambda e: str(e),
        }


class HttpError(Exception):
    def __init__(
        self,
        code: str,
        severity: Severity,
        status_code: int,
        message: LocalizedMessage,
        details: dict[Any, Any],
        debug_code: Optional[str] = None,
        debug_message: Optional[str] = None,
        occurred_while: Optional[str] = None,
        debug_details: dict[Any, Any] = {},
        caused_by: Optional[Exception] = None,
    ) -> None:
        self.public_details = HttpErrorPublicDetails(
            code=code,
            status_code=status_code,
            message=message,
            details=details,
        )

        self.debug_details = HttpErrorDebugDetails(
            severity=severity,
            debug_code=debug_code or code,
            occurred_while=occurred_while,
            caused_by=str(caused_by) if caused_by else None,
            debug_message=debug_message or message.en,
            debug_details=debug_details,
        )


class UnexpectedError(HttpError):
    def __init__(
        self,
        severity: Severity = Severity.CRITICAL,
        debug_code: Optional[str] = None,
        debug_message: Optional[str] = None,
        occurred_while: Optional[str] = None,
        debug_details: dict[Any, Any] = {},
        caused_by: Optional[Exception] = None,
    ) -> None:
        super().__init__(
            status_code=500,
            severity=severity,
            code="UNEXPECTED_ERROR",
            message=LocalizedMessage(
                ar="حدث خطأ غير متوقع. يرجى المحاولة مرة أخرى لاحقًا.",
                en="Something went wrong. Please try again later.",
            ),
            details={},
            debug_code=debug_code,
            debug_message=debug_message,
            occurred_while=occurred_while,
            debug_details=debug_details,
            caused_by=caused_by,
        )


# MARK: DEPRECATED


@deprecated("BaseError is deprecated. You should use `HttpError` instead,")
class BaseError(Exception):
    def __init__(
        self,
        status_code: int,
        occurred_while: str,
        caused_by: Optional[Exception] = None,
        debug_description: Optional[str] = None,
        message: Optional[str] = None,
    ):
        self.status_code = status_code
        self.occurrence_context = occurred_while
        self.original_error = caused_by
        self._debug_description = debug_description
        self._message = message

    @property
    def debug_description(self) -> str:
        debug_description = f"Error occurred while {self.occurrence_context}\n"

        if self._debug_description:
            debug_description += f"with message: {self._debug_description}\n"

        if self.original_error:
            debug_description += f"caused by: {self.original_error}"

        return debug_description

    @property
    def message(self) -> str:
        if self._message:
            return self._message
        return f"Something went wrong while {self.occurrence_context}"



from typing import Any, TypeVar
from core_infinity_stones.errors.base_error import BaseError

T = TypeVar("T")


class MapIsMissingKeysError(BaseError):
    def __init__(
        self,
        status_code: int,
        occurred_while: str,
        map_name: str,
        map: dict[T, Any],
        missing_keys: list[T],
    ):
        debug_description = f"""The following keys are missing from the {map_name}
        missing_keys: {missing_keys}
        map: {map}
        """

        super().__init__(
            status_code=status_code,
            occurred_while=occurred_while,
            debug_description=debug_description,
        )



from enum import StrEnum
from typing import Any, Optional

from pydantic import BaseModel

from core_infinity_stones.errors.base_error import LocalizedMessage, Severity


class TracingDetails(BaseModel):
    trace_id: str
    span_id: str


class EventType(StrEnum):
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"


class Event(BaseModel):
    code: str
    type: EventType
    service: str
    message: str
    details: Optional[dict[str, Any]] = None


class EventWithTraces(Event):
    trace_id: str
    span_id: str

    @classmethod
    def from_event(cls, event: Event, tracing_details: TracingDetails) -> "EventWithTraces":
        return cls(
            trace_id=tracing_details.trace_id,
            span_id=tracing_details.span_id,
            code=event.code,
            type=event.type,
            service=event.service,
            message=event.message,
            details=event.details,
        )


class ErrorEvent(EventWithTraces):
    severity: Severity
    occurred_while: Optional[str] = None
    caused_by: Optional[Exception]
    status_code: int
    public_code: str
    public_message: LocalizedMessage
    public_details: Optional[dict[str, Any]] = None


import time
import functools
import asyncio

from typing import Awaitable, Callable, Type, Tuple, TypeVar, ParamSpec, cast

T = TypeVar("T")
P = ParamSpec("P")


def retry(
    max_attempts: int = 3,
    delay_in_seconds: float = 1.0,
    exceptions: Tuple[Type[BaseException], ...] = (Exception,),
) -> Callable[[Callable[P, T]], Callable[P, T]]:
    """
    A retry decorator that works with both sync and async functions,
    and preserves type annotations.
    """

    def decorator(func: Callable[P, T]) -> Callable[P, T]:
        if asyncio.iscoroutinefunction(func):

            @functools.wraps(func)
            async def async_wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
                for attempt in range(1, max_attempts + 1):
                    try:
                        return await cast(Callable[P, Awaitable[T]], func)(
                            *args, **kwargs
                        )
                    except exceptions:
                        if attempt == max_attempts:
                            raise
                        await asyncio.sleep(delay_in_seconds)
                raise RuntimeError(
                    f"Function {func.__name__} failed after {max_attempts} attempts"
                )

            return cast(Callable[P, T], async_wrapper)

        else:

            @functools.wraps(func)
            def sync_wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
                for attempt in range(1, max_attempts + 1):
                    try:
                        return func(*args, **kwargs)
                    except exceptions:
                        if attempt == max_attempts:
                            raise
                        time.sleep(delay_in_seconds)
                raise RuntimeError(
                    f"Function {func.__name__} failed after {max_attempts} attempts"
                )

            return sync_wrapper

    return decorator



from typing import Any, Callable, Sequence
import xml.etree.ElementTree as ET

from core_infinity_stones.serialization.schemas import XmlEncodableModel


class XmlEncoder:
    def __init__(
        self,
        key_encoder: Callable[[str], str] = lambda key: key,
        param_value_encoder: Callable[[Any], str] = str,
    ):
        self.key_encoder = key_encoder
        self.param_value_encoder = param_value_encoder

    def encode_as_xml_element(self, model: XmlEncodableModel) -> ET.Element:
        properties_dict = model.__dict__
        xml_element_params = self._encode_params(properties_dict)

        element = ET.Element(model.xml_tag_name, xml_element_params)
        sub_elements = self._encode_sub_elements(properties_dict)

        element.extend(sub_elements)

        return element

    def encode_as_xml_string(
        self, model: XmlEncodableModel, encoding: str = "unicode"
    ) -> str:
        xml = self.encode_as_xml_element(model)
        return ET.tostring(xml, encoding=encoding)

    def _is_sequence_of_xml_encodable_models(self, value: Any) -> bool:
        return isinstance(value, list) and all(
            isinstance(item, XmlEncodableModel) for item in value
        )

    def _is_sub_element(self, value: Any) -> bool:
        return isinstance(
            value, XmlEncodableModel
        ) or self._is_sequence_of_xml_encodable_models(value)

    def _encode_params(self, properties_dict: dict[str, Any]) -> dict[str, str]:
        xml_element_params: dict[str, str] = {}

        for key, value in properties_dict.items():
            if value is None:
                continue

            if self._is_sub_element(value):
                continue

            encoded_key = self.key_encoder(key)
            encoded_value = self.param_value_encoder(value)

            xml_element_params[encoded_key] = encoded_value

        return xml_element_params

    def _encode_sub_elements(self, properties_dict: dict[str, Any]) -> list[ET.Element]:
        sub_elements = []

        for _, value in properties_dict.items():
            if value is None:
                continue

            if not self._is_sub_element(value):
                continue

            if isinstance(value, Sequence):
                for item in value:
                    if not isinstance(item, XmlEncodableModel):
                        continue
                    xml = self.encode_as_xml_element(item)
                    sub_elements.append(xml)

            elif isinstance(value, XmlEncodableModel):
                xml = self.encode_as_xml_element(value)
                sub_elements.append(xml)

        return sub_elements



from logging import Logger as NativeLogger
from typing import Callable

from core_infinity_stones.errors.base_error import HttpError
from core_infinity_stones.logging.schemas import (
    ErrorEvent,
    Event,
    EventType,
    EventWithTraces,
    TracingDetails,
)


class Logger:
    def __init__(
        self,
        service_name: str,
        tracing_details_resolver: Callable[[], TracingDetails],
        logger: NativeLogger,
    ):
        self.service_name = service_name
        self.tracing_details_resolver = tracing_details_resolver
        self.logger = logger

    def info(self, event: Event) -> None:
        tracing_details = self.tracing_details_resolver()
        self.logger.info(EventWithTraces.from_event(event, tracing_details))

    def warning(self, event: Event) -> None:
        tracing_details = self.tracing_details_resolver()
        self.logger.info(EventWithTraces.from_event(event, tracing_details))

    def error(self, error: HttpError) -> None:
        tracing_details = self.tracing_details_resolver()

        self.logger.error(
            ErrorEvent(
                trace_id=tracing_details.trace_id,
                span_id=tracing_details.span_id,
                code=error.debug_details.debug_code,
                type=EventType.ERROR,
                service=self.service_name,
                message=error.debug_details.debug_message,
                details=error.debug_details.debug_details,
                severity=error.debug_details.severity,
                occurred_while=error.debug_details.occurred_while,
                caused_by=error.debug_details.caused_by,
                status_code=error.public_details.status_code,
                public_code=error.public_details.code,
                public_message=error.public_details.message,
                public_details=error.public_details.details,
            )
        )


from typing import Callable, Optional, TypeVar

from core_infinity_stones.errors.response_status_code_error import (
    ResponseStatusCodeError,
)

from .schemas import HttpxResponse

T = TypeVar("T")


class ResponseHandler:
    @staticmethod
    def handle(
        response: HttpxResponse,
        source_url: str,
        decoder: Callable[[HttpxResponse], T] = lambda response: response.text,  # type: ignore
        successful_status_codes_range: range = range(200, 300),
        error_messages_by_status_codes: Optional[dict[int, str]] = None,
    ) -> T:
        if response.status_code in successful_status_codes_range:
            return decoder(response)

        raise ResponseStatusCodeError(
            url=source_url,
            status_code=response.status_code,
            debug_description=response.text,
            messages_by_status_codes=error_messages_by_status_codes,
        )
